//////////////////////////////////////////////////////////////////////////
//physics system by Roel Smit
//////////////////////////////////////////////////////////////////////////
//version 1.1
//////////////////////////////////////////////////////////////////////////
//features
// - physic movement
// - physic rotation
// - child&parent
// - open source
// - sounds
// - royalty free
// - realtime customizable skills and flags
// - examples
//////////////////////////////////////////////////////////////////////////
//function prototypes
function physic_move();
function physic_on_shoot();
function physic_getbounce();
function clamp_speed();
function setmovemode();
//////////////////////////////////////////////////////////////////////////
//vars
//set the floor range to a high value, so the my.floor_dist is active for 8000 quants
var floor_range = 8000;

//needed for the moving
var move_dist[3];

var move_rel_dist[3];
var move_abs_dist[3];
//the bounce angle
var bounce_pan[3];

//for tracing
var trace_from[3];
var trace_to[3];
var trace_dist = 40;

//just an test var whenever you need it
var test[3];

//we can use a sound
sound physic_hit<physic_hit.wav>;

var physic_movemode;
//////////////////////////////////////////////////////////////////////////
//the maximal gravity
define physic_gravity,100;
//////////////////////////////////////////////////////////////////////////
//skills

define turn_force,skill2;
//for a parent, to know how many childs it got,automatilcally set by it's childs
define subents,skill3;
//the weight of an entity(only used for gravity)
define mass,skill4;
//the gravity active at the moment
define gravity_active,skill5;

//if an entity hits the floor, it will bounce up,multiplied by this skill
define bounce_value,skill6;


//the maximal speed for all axisses
define max_speed,skill7;
//the active speeds
define speed_x,skill8;
define speed_y,skill9;
define speed_z,skill10;
//the Accelerate SPEED
define aspeed_x,skill11;
define aspeed_y,skill12;
define aspeed_z,skill13;
//the active ABSolute speeds
define absdist_x,skill14;
define absdist_y,skill15;
define absdist_z,skill16;
//the Accelerate ABSolute SPEED
define aabsdist_x,skill17;
define aabsdist_y,skill18;
define aabsdist_z,skill19;

define turndist_pan,skill20;
define turndist_tilt,skill21;
define turndist_roll,skill22;

define aturndist_pan,skill23;
define aturndist_tilt,skill24;
define aturndist_roll,skill25;
//at which point  is the child attached?
define attached_dist_x,skill26;
define attached_dist_y,skill27;
define attached_dist_z,skill28;
//////////////////////////////////////////////////////////////////////////
//the name says what it is
//while an entity is active, it is 'registered' to the physics engine
//to unregister the entity, just set it to 0
//to register it again just call physic_object
define active,flag1;
define no_move,flag2;
define no_rotate,flag3;
//enable to accelerate without touching the ground
define fly_force,flag4;
define _sound,flag5;
define me_parent,flag6;
///////////////////////////
//to set the positions to the positions of the parent
define attach_x,flag1;
define attach_y,flag2;
define attach_z,flag3;
//to set the angles to the angles of the parent
define attach_pan,flag4;
define attach_tilt,flag5;
define attach_roll,flag6;
//in case of a racing/car game
define front_wheels,flag7;

define disable_gravity,flag8;
//////////////////////////////////////////////////////////////////////////
//physic move. Using C_move,so its very slow
function physic_move
{

	//first the relative distance
	vec_set(move_rel_dist.x,my.speed_x);
	
	//the absolute distance
	vec_set(move_abs_dist.x,my.absdist_x);
	
	//add the gravity to the absolute distance
	move_abs_dist.z -= my.gravity_active;
	
	if(vec_dist(move_rel_dist.x,nullvector) > 0.1 || vec_dist(move_abs_dist.x,nullvector) > 0.1)
	{
		move_abs_dist.z -= my.gravity_active;
		C_move(me,move_rel_dist.x,move_abs_dist.x,physic_movemode);
	}	
}
//rotate the entity
//important! only use it right after physic_move, and don't change my_speed(except if you need to change it)
//this is also slow(c_rotate*2)
var my_angle[3];
function physic_rotate
{
	if(my.me_parent == 0)
	{
		//Maybe it doesn't work correctly everytime...
		vec_rotate(my_speed.x,my.pan);
		my_angle.tilt = my_speed.x*-my.turn_force;//replace 2 with your own values
		my_angle.roll = my_speed.y*-my.turn_force;//replace 2 with your own values
		c_rotate(me,vector(0,my_angle.tilt,my_angle.roll),glide);//let's rotate
	}
	c_rotate(me,vector(my.turndist_pan,my.turndist_tilt,my.turndist_roll),glide);//let's rotate again
}
//////////////////////////////////////////////////////////////////////////
//the events
//if there is shot at the entity, it will recieve the event_shoot and then physic_on_shoot will be executed
function physic_on_shoot
{
	//set the absdist to the difference
	//so the entity can bounce off
	vec_diff(my.absdist_x,my.x,target.x);
}
function physic_getbounce
{
	//the bounce events
	if(event_type == event_block)
	{
		//play a sound
		if(my.gravity_active > 1 && my._sound)
		{
			ent_playsound(my,physic_hit,my.gravity_active*20);
		}
		my.absdist_x = bounce.x;
		my.absdist_y = bounce.y;
		if(my.floor_dist > 60)
		{
			//but not the z
			my.absdist_z = bounce.z*my.bounce_value;
		}
		else
		{
			//let the gravity bounce
			my.gravity_active = -my.gravity_active*my.bounce_value;
		}
	}
	if(event_type == event_entity
	|| event_type == event_stuck
	|| event_type == event_push)
	{
		//set the x and y absdist to the bounce vector
		if(my.gravity_active > 1 && my._sound)
		{
			ent_playsound(my,physic_hit,my.gravity_active*20);
		}
		if(you)
		{
			your.absdist_x =my.absdist_x;
			your.absdist_y =my.absdist_y;
		}
		
		if(my.floor_dist > 10)
		{
			//but not the z
			my.absdist_z = bounce.z*my.bounce_value;
			//let the gravity bounce
			my.gravity_active = -my.gravity_active*my.bounce_value;
		}
	}
	//is there shot at the entity?
	if(event_type == event_shoot)
	{
		physic_on_shoot();
	}
}
//////////////////////////////////////////////////////////////////////////
//to set the top speed
function clamp_speed
{
	//set the maximum of speed
	my.speed_x = clamp(my.speed_x,-my.max_speed,my.max_speed);
	my.speed_y = clamp(my.speed_x,-my.max_speed,my.max_speed);
	my.speed_z = clamp(my.speed_x,-my.max_speed,my.max_speed);
	//and set the maximum of absdist
	my.absdist_x = clamp(my.absdist_x,-my.max_speed,my.max_speed);
	my.absdist_y = clamp(my.absdist_y,-my.max_speed,my.max_speed);
	my.absdist_z = clamp(my.absdist_z,-my.max_speed,my.max_speed);
}
//to slow the entity down.
//on his way your enties stop
//replace 1.03 by your own valus, but a little chance can affect all the movement!!!
function physic_slow_down
{
	my.aabsdist_x = 0;
	my.aabsdist_y = 0;
	my.aabsdist_z = 0;
	
	my.absdist_x/=1.03;
	my.absdist_y/=1.03;
	my.absdist_z/=1.03;
	
	my.aspeed_x=0;
	my.aspeed_y=0;
	my.aspeed_z=0;
	
	my.speed_x/=1.03;
	my.speed_y/=1.03;
	my.speed_z/=1.03;
	
	my.aturndist_pan = 0;
	my.aturndist_tilt = 0;
	my.aturndist_roll = 0;
	
	my.turndist_pan/=1.03;
	my.turndist_tilt/=1.03;
	my.turndist_roll/=1.03;
}
//////////////////////////////////////////////////////////////////////////
//uses: mass , gravity_active , bounce_value , lift , fly_force , max_speed ,no_move, no_rotate, speed_x,speed_y,speed_z,subents
action physic_object
{
	c_setminmax(me);
	my.active = on;
	if(!my.mass){my.mass = 50;}
	if(!my.bounce_value){my.bounce_value = 0.01;}
	if(!my.turn_force){my.turn_force = 2;}
	//if there are speed limits... limit the speed
	
	
	//make the movement polygon-precise
	my.polygon = on;
	
	//set the event
	my.enable_impact = on;
	my.enable_entity = on;
	my.enable_stuck = on;
	
	my.enable_entity = on;
	my.enable_block = on;
	my.enable_shoot = on;
	my.event = physic_getbounce;
	
	while(my.active == 1)
	{
		if(my.max_speed){clamp_speed();}
		//if the entity is on the floor or able to accelerate in air.
		if(my.floor_dist < 0 || my.fly_force == 1)
		{
			//add speed to it
			vec_add(my.speed_x,my.aspeed_x);
			vec_add(my.absdist_x,my.aabsdist_x);
		}
		vec_add(my.turndist_pan,my.aturndist_pan);
		//not on the floor?
		//move the entity down.
		
		if(my.disable_gravity == 0 && my.floor_dist > 0 && my.gravity_active < physic_gravity)
		{
			my.gravity_active += 0.02*time*my.mass;
		}
		if(my.disable_gravity == 1)
		{
			my.gravity_active = 0;
		}
		my.gravity/= (my.subents+1);
		//if the object must be moved
		if(my.no_move == 0)
		{
			//move it.
			physic_move();
		}
		//if the object must be rotated.
		if(my.no_rotate == 0)
		{
			//rotate it.
			physic_rotate();
		}
		//slowly slow it down.
		physic_slow_down();
		//wait one frame.
		wait(1);
	}
}
///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////
//to create a child object, you have to set the parent parameter to the parent.
//you can do this in the entity propeties
//uses: attach_x,attach_y,attach_z,attach_pan,attach_tilt,attach_roll,mass,bounce_value
action physic_object_c
{
	if(my.parent == null){ent_remove(me);return;}//we need a parent
	my.passable = on;
	you = my.parent;
	your.me_parent = 1;
	if(!my.mass){my.mass = 50;}
	vec_diff(my.attached_dist_x,my.x,your.x);
	//wait until all children functions are started
	wait(1);
	while(my.parent)
	{
		proc_late();
		you = my.parent;
		vec_set(temp.x,my.attached_dist_x);
		vec_rotate(temp.x,your.pan);
		if(my.attach_x){my.x = your.x;my.x+=temp.x;}
		if(my.attach_y){my.y = your.y;my.y+=temp.y;}
		if(my.attach_z){my.z = your.z;my.z+=temp.z;}
		if(my.attach_pan){my.pan = your.pan;}
		if(my.attach_tilt){my.tilt = your.tilt;}
		if(my.attach_roll){my.roll = your.roll;}
		if(my.floor_dist > 20 && my.floor_dist < 30)
		{
			your.gravity_active=-my.bounce_value*your.gravity_active;
		}
		wait(1);
	}
	ent_remove(me);
}
///////////////////////////////////////////////////////////////////
//to set a default move_mode.
//default = glide+activate_trigger+ignore_passents+ignore_passable
//you can always change it
starter setmovemode
{
	physic_movemode = glide //important
	+ activate_trigger
	+ ignore_passents
	+ ignore_passable;
}
//end of the real physics engine
///////////////////////////////////////////////////////////////////
//first of all
//the physics system uses the predefined function entity.event
//so,how do you then use your own events?
//one way, set the event to your own event and call physic_getbounce in it
//like this:
#function my_event
#{
	#	physic_getbounce();
	#  //...
#}

#action my_action
#{
	#	physic_object();
	#	my.event = my_event;//must be called AFTER physic_object()
	#	my.enable shoot = on;//for example
	#	//...
#}
//you can also use the event_type but maybe that isn't always usable
///////////////////////////////////////////////////////////////////
//examples
//example code for an enemy
action physic_enemy
{
	my.fly_force = 1;
	//using alpha as the health parameter!
	my.alpha = 100;
	my.transparent = on;
	physic_object();
	while(my.alpha > 0)
	{
		//for example,we use the player
		if(player != null)
		{
			//face the player
			vec_to_angle(test.pan,vec_diff(temp.x,player.x,my.x));
			c_rotate(me,vec_sub(test.pan,my.pan),glide);
			//if we are near...
			//and can see the player
			if(vec_dist(my.x,player.x) < 400 && trace(my.x,player.x) == 0)
			{
				//accelerate
				my.aabsdist_x = 0.1;
				//we are very near,so jump
				if(vec_dist(my.x,player.x) < 200)
				{
					my.aabsdist_z = 0.5;
				}
				//if hit...
				if(event_type == event_entity)
				{
					my.alpha -= 1;
				}
			}
		}
		//rotate the accelerating absdist
		//important! use absdist instead of the normal speed
		//if you do else, the entity won't bounce on walls
		vec_rotate(my.aabsdist_x,my.pan);
		wait(1);
	}
	//remove it
	ent_remove(me);
}

//example code for a jump&run player
//you have to add the life system to it on yourself
action physic_player
{
	//make sounds
	my._sound = 1;
	
	//don't rotate it
	my.no_rotate = on;
	
	//call it the player
	player = me;
	//set the weight
	my.mass  = 25;
	//set the bounce value.
	//by 1 the entity never stops with bouncing 
	//and if it is higher than 1 the entity will bounce higher and higher and higher.
	//so,set entity.bounce_value lower than 1 
	my.bounce_value = 0.25;
	//able to move in the air
	my.fly_force = 1;
	//make it physic
	physic_object();
	while(1)
	{
		//make it scanable
		my.enable_scan = on;
		
		//for the camera
		trace_to.x = my.x - cos(my.pan)*140;
		trace_to.y = my.y - sin(my.pan)*140;
		trace_to.z = my.z+ 80;
		trace(my.x,trace_to);
		vec_set(camera.x,target.x);
		
		//let the camera face the player
		vec_to_angle(camera.pan,vec_diff(temp.x,my.x,camera.x));
		
		//key input 
		if(key_cuu == 1)
		{
			my.aabsdist_x = 1*time;
		}
		if(key_cud == 1)
		{
			my.aabsdist_x = -1*time;
		}
		if(key_cul == 1)
		{
			my.aabsdist_y = 1*time;
		}
		if(key_cur == 1)
		{
			my.aabsdist_y = -1*time;
		}
		//for jumping
		//replace 1.5 with your own values!
		if(key_space == 1 && my.floor_dist < 50)
		{
			my.gravity_active -= 1.5*time;
		}
		//for the movement
		vec_set(test.x,my.absdist_x);
		//rotate the accelerating absdist
		//important! use absdist instead of the normal speed
		//if you do else, the entity won't bounce on walls
		vec_rotate(my.aabsdist_x,vector(camera.pan,0,0));//omly the pan
		
		
		//rotate it
		c_rotate(me,vector((key_a-key_d)*10*time,0 ,0),glide);
		wait(1);
		
	}
}
//to create a default car
//uses: mass,bounce_value,turn_force
//the code is not able to set the angle of the car to the floor's, use some code in move.wdl form line 635 to 648.
action physic_car
{
	my.no_rotate = on;
	physic_object();
	while(1)
	{
		my.turndist_pan/ = 1.01;
		if(my.floor_dist < 10)
		{
			if(key_w && my.speed_x < 10){my.speed_x += 0.2;}
			if(key_s && my.speed_x >-10){my.speed_x -= 0.2;}
			if(my.speed_x > 1 || my.speed_x < -1)
			{
				my.aturndist_pan=0;
				if(key_q){my.aturndist_pan = 0.1;}
				if(key_e){my.aturndist_pan = 0.1;}
			}
		}
		wait(1);
	}
}
//the wheels for the car
//uses: front_wheels
action physic_wheels
{
	//set the x,y,z and roll to the x,y,z and roll of the parent(car)
	my.attach_x = 1;
	my.attach_y = 1;
	my.attach_z = 1;
	my.attach_pan = 0;
	my.attach_tilt = 0;
	my.attach_roll = 1;
	physic_object_c();
	if(my.front_wheels)
	{
		while(me)
		{ 
		//set the pointer to the entity's parent
			you = my.parent;
			if(my.floor_dist < 15)
			{
				//don't get too low...
				your.disable_gravity = 1;
			}
			if(my.floor_dist > 20)
			{
				your.disable_gravity = 0;
			}
			my.pan = your.pan+your.turndist_pan*15;//to turn
			my.tilt -= your.speed_x*60;//to spin, 60 is just a value. you can set it to any value you want
			wait(1);
		}
	}
	else
	{
		my.attach_pan = 1;
		while(me)
		{ 
		//set the pointer to the entity's parent
			you = my.parent;
			if(my.floor_dist < 15)
			{
				//don't get too low...
				your.disable_gravity = 1;
			}
			if(my.floor_dist > 20)
			{
				your.disable_gravity = 0;
			}
			my.tilt -= your.speed_x*60;//to spin
			wait(1);
		}
	}
}
action car_ai
{
	physic_object();
	
	var test_1[3] = 320,-100,0;
	var test_2[3] = 320,100,0;
	while(1)
	{
		vec_set(temp.x,test_1.x);
		vec_rotate(temp.x,my.pan);
		vec_add(temp.x,my.x);
		//to scan for objects
		if(c_trace(my.x,temp.x,ignore_me) > 0)
		{
			my.aturndist_pan = 0.5;
			my.speed_x/=1.1;
		}
		vec_set(temp.x,test_2.x);
		vec_rotate(temp.x,my.pan);
		vec_add(temp.x,my.x);
		//to scan for objects
		if(c_trace(my.x,temp.x,ignore_me) > 0)
		{
			my.aturndist_pan = -0.5;
			my.speed_x/=1.1;
		}
		
		my.aturndist_pan/= 1.03;
		my.turndist_pan/= 1.03;
		my.speed_x += (10-my.speed_x)/20;
		my.turndist_pan = clamp(my.turndist_pan,-2.5,2.5);
		if(key_space){my.speed_x /=1.1;my.aturndist_pan*=2;}
		wait(1);
	}
}